/*
 * Copyright (C) 2012 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.amnesica.kryptey.inputmethod.latin;

import android.content.Context;
import android.media.AudioManager;
import android.os.Vibrator;
import android.view.HapticFeedbackConstants;
import android.view.View;

import com.amnesica.kryptey.inputmethod.latin.common.Constants;
import com.amnesica.kryptey.inputmethod.latin.settings.SettingsValues;

/**
 * This class gathers audio feedback and haptic feedback functions.
 * <p>
 * It offers a consistent and simple interface that allows LatinIME to forget about the
 * complexity of settings and the like.
 */
public final class AudioAndHapticFeedbackManager {
  private AudioManager mAudioManager;
  private Vibrator mVibrator;

  private SettingsValues mSettingsValues;
  private boolean mSoundOn;

  private static final AudioAndHapticFeedbackManager sInstance =
      new AudioAndHapticFeedbackManager();

  public static AudioAndHapticFeedbackManager getInstance() {
    return sInstance;
  }

  private AudioAndHapticFeedbackManager() {
    // Intentional empty constructor for singleton.
  }

  public static void init(final Context context) {
    sInstance.initInternal(context);
  }

  private void initInternal(final Context context) {
    mAudioManager = (AudioManager) context.getSystemService(Context.AUDIO_SERVICE);
    mVibrator = (Vibrator) context.getSystemService(Context.VIBRATOR_SERVICE);
  }

  public boolean hasVibrator() {
    return mVibrator != null && mVibrator.hasVibrator();
  }

  public void vibrate(final long milliseconds) {
    if (mVibrator == null) {
      return;
    }
    mVibrator.vibrate(milliseconds);
  }

  private boolean reevaluateIfSoundIsOn() {
    if (mSettingsValues == null || !mSettingsValues.mSoundOn || mAudioManager == null) {
      return false;
    }
    return mAudioManager.getRingerMode() == AudioManager.RINGER_MODE_NORMAL;
  }

  public void performAudioFeedback(final int code) {
    // if mAudioManager is null, we can't play a sound anyway, so return
    if (mAudioManager == null) {
      return;
    }
    if (!mSoundOn) {
      return;
    }
    final int sound;
    switch (code) {
      case Constants.CODE_DELETE:
        sound = AudioManager.FX_KEYPRESS_DELETE;
        break;
      case Constants.CODE_ENTER:
        sound = AudioManager.FX_KEYPRESS_RETURN;
        break;
      case Constants.CODE_SPACE:
        sound = AudioManager.FX_KEYPRESS_SPACEBAR;
        break;
      default:
        sound = AudioManager.FX_KEYPRESS_STANDARD;
        break;
    }
    mAudioManager.playSoundEffect(sound, mSettingsValues.mKeypressSoundVolume);
  }

  public void performHapticFeedback(final View viewToPerformHapticFeedbackOn) {
    if (!mSettingsValues.mVibrateOn) {
      return;
    }
    if (mSettingsValues.mKeypressVibrationDuration >= 0) {
      vibrate(mSettingsValues.mKeypressVibrationDuration);
      return;
    }
    // Go ahead with the system default
    if (viewToPerformHapticFeedbackOn != null) {
      viewToPerformHapticFeedbackOn.performHapticFeedback(
          HapticFeedbackConstants.KEYBOARD_TAP,
          HapticFeedbackConstants.FLAG_IGNORE_GLOBAL_SETTING);
    }
  }

  public void onSettingsChanged(final SettingsValues settingsValues) {
    mSettingsValues = settingsValues;
    mSoundOn = reevaluateIfSoundIsOn();
  }

  public void onRingerModeChanged() {
    mSoundOn = reevaluateIfSoundIsOn();
  }
}
